open Printf
open Ext
open Slearn


(* Data structure for a set of integers *)
module IntSet =
struct
  include Hashtbl

  let add h x = add h x ()
  let iter f h = iter (fun x () -> f x) h
  let fold f h = fold (fun x () l -> f x l) h
  let to_list h = fold (fun x l -> x :: l) h []
  let sum_map f h = fold (fun x accu -> f x + accu) h 0
  let sumf_map f h = fold (fun x accu -> f x +. accu) h 0.0
  let filter f h = iter (fun n -> if not (f n) then remove h n) h
end

(* Values of the per-split prior.  If the -psthresh option is used,
 * then we save the current circuit every time we fall below a new
 * threshold.
 *)
let prior_threshes = 
 [1000.0; 500.0; 200.0; 100.0; 50.0; 20.0; 10.0; 5.0; 2.0; 1.0; 
 0.5; 0.2; 0.1; 0.05; 0.02; 0.01; 0.0]

let prior_thresh_names = 
  ["1e3"; "5e2"; "2e2"; "1e2"; "5e1"; "2e1"; "1e1"; "5e0"; "2e0"; "1e0";
  "5em1"; "2em1"; "1em1"; "5em2"; "2em2"; "1em2"; "zero"]
  
let curr_thresh = ref prior_threshes
let curr_tname  = ref prior_thresh_names

let big_int = 1000000000 (* = 1 billion *)

(* Globals used for command line parameters *)

let datafile = ref ""

(* The schema file allows you to specify the dimension of each
   variable.  Its format is a single line containing each dimension,
   separated by spaces, surrounded by parentheses, e.g.:
(2 2 3 2 5)

   For a domain where the first two variables have values 0 and 1, the
   third has values 0, 1, and 2, etc.
 *)
let schemafile = ref ""

(* Output arithmetic circuit *)
let outfile = ref ""

(* Output Bayesian network (xmod format) *)
let bnoutfile = ref ""

(* The weight the score function assigns to each additional edge in the AC. *)
let edge_cost = ref 0.1

(* The weight the score function assigns to each additional split created,
   i.e., the penalty for each additional parameter. *)
let split_cost = ref (-1.)

(* Maximum number of edges and splits allowed.  Sometimes these are
   exceeded a little bit, due to imprecise coding. *)
let max_edges = ref big_int
let max_splits = ref big_int

(* After every k splits, write the current circuit to disk.  If less
   than 0, skip this. *)
let write_freq = ref (-1)

(* Output models for different per-split thresholds.  This is good for
 * tuning the per-split penalty in a single training pass, rather than
 * n separate training runs.
 *)
let thresh = ref false

(* Use quick heuristic (default is greedy) *)
let quick = ref false

(* After converging with quick heuristic, switch to greedy to see if
   we can further improve the model. *)
let do_greedy = ref false

(* If we converge with one per-edge cost and shrink_edge_cost is true,
 * we will reduce the per-edge cost and keep going.  This is good for
 * when you have a fixed edge limit you're trying to meet.
 *)
let shrink_edge_cost = ref false

(* Bayes net file to use as prior info. *)
let bnfile = ref ""

(* Limit parents of each variable to be parents in the BN... *)
let limitparents = ref false

(* ...or ancestors in the BN... *)
let limitanc = ref false

(* Skip learning an AC, and just learn a BN (similar to WinMine) *)
let no_ac = ref false

(* Allow cycles *)
let depnet = ref false


let usage = "Usage: aclearnstruct -i <input> -o <output> [...]"
let args = Arg.align
 ([("-i", Arg.Set_string datafile, " Training data file") ;
   ("-s", Arg.Set_string schemafile, " Data schema (optional)") ;
   ("-o", Arg.Set_string outfile, " Output circuit") ;
     (* The two options below are currently HACKS *)
   ("-noac", Arg.Set no_ac, " Only learn a BN, not an AC (faster)") ;
   ("-depnet", Arg.Set depnet, " Allow cycles") ;
     (* This also implies a per-edge cost of 0. *)
   ("-ob", Arg.Set_string bnoutfile, " Output Bayesian network (XMOD format)") ;
   ("-bn", Arg.Set_string bnfile, " BN to use for partial order") ;
   ("-limitanc", Arg.Set limitanc, " Restrict parents to ancestors in the BN") ;
   ("-limitparents", Arg.Set limitparents, " Restrict parents to those listed in the BN") ;
   ("-pe", Arg.Set_float edge_cost, " Per-edge penalty") ;
   ("-shrink", Arg.Set shrink_edge_cost, " Shrink edge cost before halting") ;
   ("-ps", Arg.Set_float split_cost, " Per-split penalty") ;
   ("-psthresh", Arg.Set thresh, 
     " Output models for various per-split penalty thresholds");
   ("-maxe", Arg.Set_int max_edges, " Maximum number of edges") ;
   ("-maxs", Arg.Set_int max_splits, " Maximum number of splits") ; 
   ("-quick", Arg.Set quick, " 'Quick and dirty' heuristic") ;
   ("-qgreedy", Arg.Set do_greedy, " Find greedy local optimum after quick heuristic") ;
   ("-freq", Arg.Set_int write_freq, 
     " Number of iterations between saving current circuit")]
   @ common_arguments)
  
  (*
let output_feature_conds stream addl = 
  let fhash = Hashtbl.create 100 in
  List.iter (fun add -> feature_conds_rec fhash add.root []) addl;
  let print_condl condl =
    List.iter (fun (var, value) -> fprintf stream " v%d=%d " var value) condl;
    fprintf stream "\n" in
  let print_id_conds id condll =
    fprintf stream "ID %d\n" id;
    List.iter print_condl condll in
  Hashtbl.iter print_id_conds fhash
  *)

let rec update_tree_rec a currnode = function
| [] -> 
  (* Base case: If no conditions remain, construct a leaf. *)
  (match currnode with Bn.Vertex _ -> assert false | Bn.Leaf _ -> ());
  Bn.Leaf a
| (sense,var,value) :: condl ->
  (* Recursive case: Construct a vertex for this condition
     (perhaps based on an existing vertex) *)
  begin
    match currnode with 
      Bn.Vertex(var', value', left, right) ->
        (* The ordering of the conditions *must* match the ordering of
         * the splits in the current tree, or we fail. *)
        assert(var = var');
        if (value = value') = sense then
          let left' = update_tree_rec a left condl in
          Bn.Vertex(var', value', left', right)
        else
          let right' = update_tree_rec a right condl in
          Bn.Vertex(var', value', left, right')
    | Bn.Leaf [||] ->
        let child = update_tree_rec a (Bn.Leaf [||]) condl in
        let other = Bn.Leaf [||] in
        if sense then
          Bn.Vertex(var, value, child, other)
        else
          Bn.Vertex(var, value, other, child)
    | Bn.Leaf _ -> 
       (* We should never get here -- this indicates that a leaf has 
        * appeared twice *)
        assert false
  end


(* Create a BN with tree CPDs from the dists *)
let dists_to_bn schema dists =

  (* Initialize empty BN *)
  let numvars = Array.length schema in
  let bn = Bn.create_empty_network schema in
  let trees = Array.create numvars (Bn.Leaf [||]) in

  (* Build CPDs by turning each dist into a leaf *)
  let update_cpd dist =
    let a = Array.of_list (List.map Circuit.const_value dist.nodes) in
    let v = dist.dvar in
    trees.(v) <- update_tree_rec a trees.(v) (List.rev dist.cond) in
  List.iter update_cpd dists;
  for i = 0 to numvars - 1 do
    bn.Bn.dists.(i) <- Bn.Tree trees.(i)
  done;

  (* Build parent lists for each dist *)
  let rec make_parents pset = function
    | Bn.Leaf _ -> ()
    | Bn.Vertex(var, value, left, right) ->
        if not (IntSet.mem pset var) then
          IntSet.add pset var;
        make_parents pset left;
        make_parents pset right in
  for i = 0 to numvars - 1 do
    let p = IntSet.create 100 in
    make_parents p trees.(i);
    bn.Bn.parents.(i) <- List.rev (IntSet.to_list p)
  done;
  Bn.update_children_and_topo_vars bn;
  bn
  

let write_circuit filename circ dist_list =
  let out = open_out filename in
  let fl = List.flatten (List.map dist_to_features dist_list) in
  Circuit.output_with_features out circ fl;
  close_out out 

let write_bn filename circ dist_list =
  let out = open_out filename in
  let bn = dists_to_bn (Circuit.schema circ) dist_list in
  Bn.output_xmod out bn;
  close_out out

(* output_string out "\nDistribution list:\n";
  let dvar_cmp d1 d2 = compare d1.dvar d2.dvar in
  let by_var = List.sort dvar_cmp dist_list in
  List.iter (Slearn.output_dist out) by_var ; 
  *)

let save_circuit circ dist_list = 
  if !outfile <> "" then
    write_circuit !outfile circ dist_list;
  if !bnoutfile <> "" then
    write_bn !bnoutfile circ dist_list

let mkdir () =
  if !outfile <> "" then
  try Unix.mkdir (sprintf "%s.d" !outfile) 0o777
  with Unix.Unix_error _ -> ()

let save_periodic circ dist_list i =
  if !outfile <> "" then
    let filename = sprintf "%s.d/%d.ac" !outfile i in
    write_circuit filename circ dist_list 


let rawscore l e =
  let norml = l -. !split_cost in
  let norme = float_of_int e in
  norml -. (norme *. !edge_cost)

let score split =
  let e = match split.e with None -> 100_000_000 (* big_int *) | Some e -> e in
  let l = match split.l with None -> 0. | Some l -> l in
  rawscore l e

let mscore split = 
  let e = match split.e with None -> split.mine | Some e -> e in
  let l = match split.l with None -> split.maxl | Some l -> l in
  rawscore l e

let breakeven_edges split =
  let frac_e = (split.maxl -. !split_cost) /. !edge_cost in
  1 + int_of_float frac_e


exception MaxEdges
exception Converged

(*
let rec transclosure_rec varparents x =
  List.iter (transclosure_rec varparents) varparents.(x) ;
  let add y =
    if not (List.mem y varparents.(x)) then
      varparents.(x) <- y :: varparents.(x) in
  List.iter (fun p -> List.iter add varparents.(p)) varparents.(x) 
 *)

(* For each node, all parents of parents (grandparents) to
 * the list of parents.  By doing this operation on parents first,
 * we ensure that each node has a list of all its ancestors.
 *)
let rec transclosure_rec visited varparents x =
  if not visited.(x) then begin
    IntSet.iter (transclosure_rec visited varparents) varparents.(x) ;
    let add y =
      if not (IntSet.mem varparents.(x) y) then
        IntSet.add varparents.(x) y in
    IntSet.iter (fun p -> IntSet.iter add varparents.(p)) varparents.(x) ;
    visited.(x) <- true
  end

let transclosure varparents =
  let visited = Array.create (Array.length varparents) false in
  for i = 0 to Array.length varparents - 1 do
    transclosure_rec visited varparents i 
  done
  
let is_ancestor varparents anc desc =
  anc = desc || IntSet.mem varparents.(desc) anc

(* Helper functions for maintaining a set of candidate splits *)
module SplitSet = Slearn.SplitSet
let set_create () = ref SplitSet.empty
let set_add x s = s := SplitSet.add x !s
let set_rem x s = s := SplitSet.remove x !s
let set_addl xl s = List.iter (fun x -> set_add x s) xl
let set_filter f s = s := SplitSet.filter f !s
let set_iter f s = SplitSet.iter f !s
let set_elements s = SplitSet.elements !s
let set_best s = SplitSet.min_elt !s
let set_is_empty s = SplitSet.is_empty !s

(* Use a sorted list (Set) to keep track of our maxscore for each
 * split.  When the min number of edges changes, we need to update 
 * our maxscore (and reinsert the split in the split set).
 *)

let rescored_split q x =
  let mine = 
    if q then (match x.e with None -> x.mine | Some e -> e)
    else 0 in
  let maxl = match x.l with None -> x.maxl | Some l -> l in
  let x' = {d=x.d ; v=x.v ; l=x.l ; maxl=maxl ; e=x.e ; mine=mine ;
            maxscore=mscore x ; vnodes=x.vnodes ; hashid = x.hashid} in
  x'

let resort_splits split_set all_splits =
  let check_split s =
    if mscore s <> s.maxscore then
      (set_rem s split_set; set_add (rescored_split !quick s) split_set) in
  List.iter check_split all_splits


let rec best_split circ anc maxe_remaining max_maxe splitset seen = 

  (* Get best split *)
  let x = set_best splitset in

  (* If we've seen this before, then it really is the best!
     (NOTE: We can sometimes skip this for greedy, but at worst 
     it's one extra split evaluation per iteration.) *)
  if !no_ac || SplitHash.mem seen x then (x, max_maxe)
  (* Otherwise, recompute its edge cost, reinsert it with a new
     maxscore, and repeat. *)
  else begin
    (* Compute the change in number of edges, if necessary *)
    (* Stop once our score becomes less than zero *)
    let max_edge_delta = min (breakeven_edges x) maxe_remaining in
    if max_edge_delta < 0 then (x, max_maxe)
    else begin
      let e = Slearn.delta_edges circ anc max_edge_delta x in 
      (* Keep track of our worst case *)
      let max_maxe = max e max_maxe in
      x.e <- Some e;
      set_rem x splitset;
      set_add (rescored_split !quick x) splitset;
      SplitHash.add seen x ();
      best_split circ anc maxe_remaining max_maxe splitset seen
    end
  end


let rec pick_best_split circ anc splitset dist_list =
  if set_is_empty splitset then 
    (vlogf "No splits remaining!"; raise Converged)
  else begin
  let maxe_remaining = !max_edges - Circuit.num_edges circ in
  let visited_splits = SplitHash.create 100 in
  let (best, max_maxe) = 
    best_split circ anc maxe_remaining 0 splitset visited_splits in
  nlogf "Splits recomputed: %d ; worst-case edges: %d\n" 
    (SplitHash.length visited_splits) max_maxe;
  vlogf "Best score: %f\n" (score best);
  if score best < 0.0 then 
    if !do_greedy then begin
      nlogf "Quick learning converged; saving circuit and switching to greedy.\n";
      save_circuit circ dist_list;
      (* Change the output filename, so we keep the quick version separate. *)
      if !outfile <> "" then
        outfile := sprintf "%sg" !outfile;
      (* Switch over to fully greedy operation. *)
      quick := false ;
      let splitlist = set_elements splitset in
      List.iter (fun s -> s.e <- None) splitlist ;
      resort_splits splitset splitlist;
      pick_best_split circ anc splitset dist_list
    end
    else if !shrink_edge_cost && maxe_remaining > 0 
          (* Don't keep going if we have less than 1% of the total edges
             remaining. *)
          && !max_edges / maxe_remaining < 100 then
        begin
          nlogf "Learning converged with edge penalty of %f.\n" !edge_cost;
          edge_cost := !edge_cost /. 2.;
          nlogf "Continuing with new edge cost of %f.\n" !edge_cost;
          save_circuit circ dist_list;
          (* Our split cost estimates might be off, since they were
             upper bounded with a different edge cost. *)
          let splitlist = set_elements splitset in
          List.iter (fun s -> s.e <- None) splitlist ; 
          resort_splits splitset splitlist;
          pick_best_split circ anc splitset dist_list
        end
    else raise Converged 
  else best
  end


let load_data filename =
  Data.input_example_list (open_in filename)

let loglikelihood circ data =
  let scratch = Circuit.create_scratch circ in
  Array.sumf_map (Circuit.logprob_x scratch circ) data 

let do_learn () =
  vlogf "Initializing...\n" ;

  (* Read in data and determine schema (number of values for each var) *)
  let data = load_data !datafile in 
  vlogf "Loaded data.\n";
  let schema = 
    if !schemafile <> "" then begin
      let schemain = open_in !schemafile in
      let s = Data.input_example schemain in
      close_in schemain ; s
    end
    else Data.schema data in 
  let data = Array.of_list data in

  (* Create initial circuit and set of distributions. (Initial
     distributions will be single-variable marginals.) *)
  let (circ, initial_dists, split_list) = 
    Slearn.init schema data in
  let dist_list = ref initial_dists in

  (* Load a BN, to determine a partial order to follow when
     learning our circuit. *)
  let bn =
    if !bnfile = "" then Bn.create_empty_network schema
    else Bn.load_auto !bnfile in

  (* Create varparents, an array of IntSets, and optionally initialize
     it using the BN parents. *)
  let init v = 
    let is = IntSet.create 10 in
    if !bnfile <> "" then 
    List.iter (IntSet.add is) (Bn.parents bn v); 
    is in
  let varparents = Array.init (Array.length (Circuit.all_vars circ)) init in

  (* If we're limiting the parents to be those in the BN,
   * then we must make a distinction between parents and ancestors.
   * Therefore, we should not take the transitive closure.
   *)
  if not !limitparents then
    transclosure varparents;

  (* Create a set of candidate splits *)
  let split_set = set_create () in
  set_addl split_list split_set;

  (* DEBUG: Check that every split in our list made it into the set.
   * Our use of the Set data structure is convenient for maintaining
   * order, but if two splits ever have the exact same delta log
   * likelihood, then only one will remain in the Set.
   *)
  let initial_num_splits = List.length split_list in
  let num_splits () = List.length (set_elements split_set) in
  if initial_num_splits <> num_splits() then begin
    nlogf "ERROR: Number of splits decreased from %d to %d!\n"
      initial_num_splits (num_splits ());
    nlogf "This means that %d splits were erroneously rejected as duplicates.\n"
      (initial_num_splits - num_splits ());
  end;
  vlogf "Initial splits: %d\n" (num_splits ());

  (* Filter out illegal splits *)
  let valid s = 
    let dvar = s.d.dvar and vvar = (Circuit.var_var s.v) in
    if !limitparents || !limitanc then
      IntSet.mem varparents.(dvar) vvar
    else 
      not (is_ancestor varparents dvar vvar) in
  set_filter valid split_set ;
  vlogf "Splits after filtering: %d\n" (num_splits ());

  (* Initial log-likelihood *)
  let init_ll = loglikelihood circ data in
  if !write_freq > 0 || !thresh then mkdir () ;
  nlogf "Initializing took %f seconds.\n" (Sys.time ());
  nlogf "Initial log likelihood: %f\n" init_ll;
  let lasttime = ref (Sys.time()) in

  (* Choose best split, one at a time.  If no split increases the
     score function, then we leave this loop early by raising an
     exception. *)
  (try
  let i = ref 0 in
  while !i < !max_splits do
    incr i ;
    nlogf "\nSPLIT %d:\n" !i;
    (* DEBUG: print out z, to detect some types of circuit corruption. *)
    if log_exists log_debug then begin
      let scratch = Circuit.create_scratch circ in
      dlogf "z = %f\n" (Circuit.compute_z scratch circ) 
    end;
    let beg_time = Sys.time() in
    let anc = Slearn.build_ancestors circ in
    let split = pick_best_split circ anc split_set !dist_list in

    (* Save periodic circuits representing where we'd stop if
     * we had various values of the per-split penalty, -ps
     * Only active if running with -psthresh
     *)
    while !thresh && score split < List.hd !curr_thresh do
      assert(!outfile <> "");
      let fname = sprintf "%s.d/%s-ps%s" 
        !outfile !outfile (List.hd !curr_tname) in
      nlogf "Passed -ps threshold %f; saving circuit \"%s\".\n"
        (List.hd !curr_thresh) fname ;
      write_circuit fname circ !dist_list ;
      curr_tname := List.tl !curr_tname ;
      curr_thresh := List.tl !curr_thresh ;
    done ;
    let picked_time = Sys.time() in 
    let delta_edge = match split.e with Some e -> e | None -> -1 in
    let curr_edges = Circuit.num_edges circ in
    if delta_edge + curr_edges > !max_edges then
      raise MaxEdges ;

    (* Apply the chosen split. *)
    let all_splits = if !quick || !no_ac then [] 
                     else (set_elements split_set) in
    let d = split.d and v = split.v in
    let (d1, d2) = 
      if !no_ac then 
        Slearn.split_dist_params d v data 
      else
        Slearn.split_dist circ anc all_splits split data in
    let split_time = Sys.time() in
    let new_edges = Circuit.num_edges circ in

    (* Fix sorting of candidate splits *)
    resort_splits split_set all_splits;

    if log_exists log_verbose then begin
       vlogf "Splitting distribution:\n" ;
       Slearn.output_dist (log_stream log_verbose) split.d ;
       vlogf "Splitting on var %s\n" (Circuit.string_of_node split.v)
    end;
    nlogf "Delta edges: %d ; ll: %f\n" delta_edge split.maxl;
    (* Check that the predicted delta_edges is correct. *)
    assert(!no_ac || (new_edges - curr_edges) = delta_edge);
    nlogf "Total edges: %d ; Depth: %d\n" 
      new_edges (Circuit.depth circ);
    nlogf "Elapsed time: %f seconds\n" (Sys.time ());
    nlogf "Delta time: %f seconds\n" (Sys.time () -. !lasttime);
    lasttime := Sys.time();

    if !write_freq > 0 && !i mod !write_freq = 0 then 
      save_periodic circ !dist_list !i;

    (* Update distribution list *)
    dist_list := d1 :: d2 :: List.filter (( != ) d) !dist_list ;

    (* Remove old splits *)
    set_filter (fun s-> s.d != d) split_set ;

    (* Update parent list *)
    let dvar = Slearn.dvar d and splitvar = Circuit.var_var v in
    if not !depnet && not (IntSet.mem varparents.(dvar) splitvar) then begin
      IntSet.add varparents.(dvar) splitvar ;
      (* if not !limitparents && not !limitanc then *)
      transclosure varparents ;
      (* Remove illegal splits *)
      let valid s = 
        let dvar = s.d.dvar and vvar = (Circuit.var_var s.v) in
        not (is_ancestor varparents dvar vvar) in
      set_filter valid split_set 
    end ;
    
    (* Add new splits *)
    let split_filter d v = 
      not (is_ancestor varparents d v) in
    let dll = Slearn.gen_split circ data split_filter in
    set_addl (dll d1) split_set ;
    set_addl (dll d2) split_set ; 

    vlogf "Total number of potential splits: %d\n" (num_splits ());

    let update_time = Sys.time () in
    nlogf "Times: pick=%.4f split=%.4f update=%.4f\n"
           (picked_time -. beg_time) (split_time -. picked_time)
           (update_time -. split_time);
  done ;
  nlogf "\nMaximum number of splits reached.\n" ;
  with Converged -> nlogf "\nLearning converged.\n" 
     | MaxEdges -> nlogf "\nMaximum number of edges reached.\n" );
  save_circuit circ !dist_list;
  (* Compute the final log likelihood.  Doing it this way will
     always be correct, but could be rather slow... *)
  if not !no_ac && log_exists log_debug then
    dlogf "Final log likelihood: %f\n" (loglikelihood circ data); 
  nlogf "Total time: %f seconds\n" (Sys.time ())


let main () = 
  Arg.parse args ignore usage;
  if !datafile = "" 
    || (!no_ac && !bnoutfile = "") 
    || (not !no_ac && !outfile = "") then
    (Arg.usage args usage; exit (-1));
  if !no_ac then begin
    (* Outputting models at different per-split thresholds
       is not (yet) supported for pure BN learning. *)
    thresh := false;
    edge_cost := 0.;
  end;
  (* Set default split cost *)
  if !split_cost < 0. then
    split_cost := if !no_ac then 0.1 else 0.0;
  common_log_init ();
  do_learn ()

let _ = main ()
