open Printf

(* FORMAT:

   Data examples are one per line, consisting of a comma-separated
   list of variable values.  Asterisks [*] indicate unknown values.
   e.g.
   
0,0,1,0,3
1,*,*,1,2

etc. *)

(*
 * Input 
 *)
exception Eof

(*
let example_stream channel = Lexing.from_channel channel

let input_example lexbuf =
  try DataParser.intset DataLexer.lexer lexbuf
  with DataLexer.Eof -> raise Eof;;
 *)

let comma = Str.regexp "[ \t]*,[ \t]*"

let input_example channel =
  try
    let s = input_line channel in
    let tokens = Str.split comma s in
    let vals = List.map (function "*" -> (-1) | t -> int_of_string t) tokens in
    Array.of_list vals
  with End_of_file -> raise Eof
     | Failure _ ->
        printf "ERROR: Token was neither an integer nor * in input_example.\n";
        [||]

let input_example_list channel =
  let l = ref [] in
  try
    while true do
      let x = input_example channel in
      l := x :: !l
    done; List.rev !l
  with Eof -> List.rev !l

let input_marginals channel =
  let re = Str.regexp "[ \t]" in
  let lines = ref [] in
  let linenum = ref 0 in
  try 
    while true do 
      incr linenum;
      let currline = input_line channel in
      (* TODO: Support multiple marginals in one file, eventually.
      if currline = "END" then raise End_of_file; *)
      let probstrs = Str.split re currline in
      let probs = Array.of_list (List.map float_of_string probstrs) in
      lines := probs :: !lines
    done; [||]
   with End_of_file -> Array.of_list (List.rev !lines)
      | Failure _ -> 
          printf "ERROR: Non-float marginal value read on line %d\n" !linenum; 
          [||]

(* Function for simple input validation *)
exception Invalid_length
exception Invalid_value
let check_point schema example =
  if Array.length schema != Array.length example then
    raise Invalid_length;
  for i = 0 to Array.length schema - 1 do
    if example.(i) >= schema.(i) then raise Invalid_value 
  done

let check_evidence schema ev =
  if ev = [||] then
    Array.create (Array.length schema) (-1)
  else
    (check_point schema ev; ev)

(*
 * Output
 *)

let output o x =
  for i = 0 to Array.length x - 1 do
    if i > 0 then fprintf o ",";
    if x.(i) < 0 then 
      output_string o "*" 
    else 
      output_string o (string_of_int x.(i))
  done;
  output_string o "\n"

let print = output stdout

let to_string x =
  let string_of_value xi =
    if xi < 0 then "*"
    else string_of_int xi in
  let elements = List.map string_of_value (Array.to_list x) in
  (String.concat "," elements) ^ "\n"

let output_marginals o m =
  for i = 0 to Array.length m - 1 do 
    for j = 0 to Array.length m.(i) - 1 do
      if j > 0 then fprintf o " ";
      output_string o (string_of_float m.(i).(j))
    done;
    output_string o "\n"
  done
 (* TODO: Support multiple marginals in one file, eventually.
  ; output_string o "END\n" *)

let print_marginals o = output_marginals stdout


(*
 * Analysis
 *)

(* Copied from extensions.ml *)
let array_map2 f a1 a2 = Array.mapi (fun i x -> f x a2.(i)) a1

let data_max = function
| x :: l -> List.fold_left (array_map2 max) x l
| [] -> [||]

let schema example_list = Array.map (( + ) 1) (data_max example_list)

let stream_schema lexbuf =
  let a = input_example lexbuf in
  (try while true do
    let a' = input_example lexbuf in
    for i = 0 to Array.length a - 1 do
      a.(i) <- max a.(i) a'.(i)
    done
  done with Eof -> ());
  Array.map (( + ) 1) a
