open Printf
open Ext

let qfile = ref ""
let evfile = ref ""
let modelfile = ref ""
let marg_outfile = ref ""
let sameev = ref false
let max_iter = ref 50
let threshold = ref 0.0001

let usage = "Usage: bp -m <model> -mo <output> -ev <evidence> [...]"
let args = Arg.align
 ([("-m", Arg.Set_string modelfile, " Model file (XMOD, BIF, or MN)") ;
   ("-mo", Arg.Set_string marg_outfile, " Output marginals") ;
   ("-ev", Arg.Set_string evfile, " Evidence file");
   ("-sameev", Arg.Set sameev, " Use the same evidence for all queries") ;
   ("-q",  Arg.Set_string qfile,  " Query file");
   ("-maxiter", Arg.Set_int max_iter, " Maximum number of iterations [50]") ;
   ("-thresh", Arg.Set_float threshold, " Convergence threshold [0.0001]")]
   @ common_arguments)

let load_mn file =
  match filetype file with
  | MNFile -> Mn.load (open_in file)
  | BNFile -> Bn.to_mn (Bn.load_auto file)
  | _ -> nlogf "ERROR: Unsupported model file type.\n"; exit(-1)

(* HACK: Stolen from mf.ml... TODO: Refactor! *)
let condition_marginals ev marginals =
  if Array.length ev = 0 then marginals
  else begin
    let cond_one e m = 
      if e < 0 then m
      else begin
        let m' = Array.create (Array.length m) 0. in
        m'.(e) <- 1.;
        m'
      end in
    Array.map2 cond_one ev marginals
  end

let do_approx () =
  let mn = load_mn !modelfile in
  let numvars = Mn.numvars mn in
  let schema = Mn.schema mn in

  (* Process arguments, open streams *)
  let evstream = if !evfile <> "" then open_in !evfile else stdin in 
  let qstream = if !qfile <> "" then open_in !qfile else stdin in 
  if !evfile = "" then sameev := true;
  let sharedmarg = 
    if !sameev then begin
      start_time "bp";
      let ev = if !evfile = "" then [||] else Data.input_example evstream in
      let ev = Data.check_evidence schema ev in
      let sm = BpImpl.run_bp !max_iter !threshold mn ev in
      vlogf "BF time: %fs\n" (elapsed "bp");
      (ev, sm)
    end else ([||], [||]) in
  let mout = if !marg_outfile <> "" then open_out !marg_outfile 
             else log_stream log_normal in

  let first_marginal = ref true in
  let sl = stats_make () in
  let st = stats_make () in
  begin
  try while true do
    start_time "query";
    let (ev, marg) = 
      if !sameev then
        sharedmarg 
      else
        let ev = Data.input_example evstream in
        let ev = Data.check_evidence schema ev in
        (ev, BpImpl.run_bp !max_iter !threshold mn ev) in

    (* Write out marginals *)
    if !marg_outfile <> "" || !qfile = "" then begin
      (* Separate two marginal sets with a blank line *)
      if !first_marginal then
        first_marginal := false
      else
        output_string mout "\n";
      let marg = Array.map (Array.map exp) marg in
      Data.output_marginals mout (condition_marginals ev marg)
    end;

    if !qfile <> "" then begin
      let q = Data.input_example qstream in
      if (Array.length q) != numvars then
       (printf "Error: Query is wrong size or format!\n";
        flush stdout; assert false);
      let llvar var vmarg =
        if q.(var) >= 0 && ev.(var) < 0 then vmarg.(q.(var)) else 0. in
      let l = Array.sumf (Array.mapi llvar marg) in
      let dt = elapsed "query" in
      if log_exists log_verbose then
        vlogf "%f %f\n" l dt
      else
        nlogf "%f\n" l;
      stats_add sl l;
      stats_add st dt
    end;
    (* Stop immediately if we only have a single piece of 
     * evidence and no queries. *)
    if !sameev && !qfile = "" then raise Data.Eof
  done with Data.Eof -> ()
  end;
  (sl, st)


let main () = 
  start_time "total";
  Arg.parse args ignore usage;
  if !modelfile = "" then
    (Arg.usage args usage; exit (-1));

  common_log_init ();
  let (sl, st) = do_approx () in

  (* Print stats *)
  if stats_n sl > 0 then begin
    if classify_float (stats_stderr st) = FP_nan then
      vlogf "time = %f +/- 0\n" (stats_mean st)
    else
      vlogf "time = %f +/- %f\n" (stats_mean st) (stats_stderr st);
    nlogf "avg = %f +/- %f\n" (stats_mean sl) (stats_stderr sl);
  end;
  vlogf "Total time: %fs\n" (elapsed "total")

;;
main ()
