open Ext

(* Explanation of common variables:
 *
 * neighbors -- mapping from neighbor index to variable number
 *   e.g., if a potential has neighbors 5, 1, 2, 3, then
 *   this might be the array: [|5;1;2;3|]
 *
 * nmap -- mapping from variables to neighbor indices.
 *   e.g., if a potential has neighbors 5, 1, 2, 3, then
 *   this might be the array: [|-1;1;2;3;-1;0|], where
 *   "-1" denotes "not a neighbor".
 *
 * imsg -- array of incoming messages to the factor, 
 *   indexed by neighbor index.  Each is stored in log
 *   space and normalized so that:
 *     forall_i (sum_j exp imsg.(i).(j) = 1).
 *
 * omsg -- Same as imsg, but outgoing messages
 *)

(* Set a variable in conditions array to either value
   or !value, depending on sense (which is true or false).
   Returns a function that resets this change. *)
let set_condition conditions (sense, var, value) =
  if sense then begin
    (* Save original states *)
    let old = Array.copy conditions.(var) in
    (* Set conflicting values to false *)
    for i = 0 to Array.length conditions.(var) - 1 do
      if i <> value then
        conditions.(var).(i) <- false
    done;
    (* Construct reset function to restore states *)
    let reset () =
      for i = 0 to Array.length conditions.(var) - 1 do
        conditions.(var).(i) <- old.(i)
      done in
    reset
  end else begin
    (* Save original state *)
    let orig = conditions.(var).(value) in
    (* Set specified value to false *)
    conditions.(var).(value) <- false;
    (* Construct reset function *)
    let reset () = conditions.(var).(value) <- orig in
    reset
  end

module F = Mn.Factor

let rec add_probs schema imsg omsg neighbors nmap accu conditions = function 
(* If it's a leaf, add appropriate probability to all compatible slots *)
| F.Leaf wt ->
    let accu' = accu +. wt in
    if accu' <> log_zero then begin
      (* Add the corresponding probabilities to all compatible marginals *)
      (* var is the variable; i is its index in the arrays of
         incoming and outgoing messages. *)
      let add_marginals i var =
        (* Compute total prob. of remaining values for ith variable,
           var.  If some of them were excluded above (i.e., conditions
           was set to false, then the probability of those cases
           has already been factored into accu. *)
        let total = ref log_zero in
        for value = 0 to schema.(var) - 1 do
          if conditions.(var).(value) then
            total := logsumexp2 !total imsg.(i).(value)
        done;
        (* Add actual probabilities *)
        for value = 0 to schema.(var) - 1 do
          if conditions.(var).(value) then
            omsg.(i).(value) <- logsumexp2 omsg.(i).(value) 
              (accu' +. imsg.(i).(value) -. !total)
        done in
      Array.iteri add_marginals neighbors
    end

(* Otherwise, condition on the splitting variable and recurse *) 
| F.Vertex (var, value, left, right) ->
    (* Compute total probability of remaining choices for var,
       in case some have already been excluded by previous splits. *)
    let total = ref log_zero in
    let subtotal = ref log_zero in (* Total marginal of other values *)
    for i = 0 to schema.(var) - 1 do
      if conditions.(var).(i) then begin
        total := logsumexp2 !total imsg.(nmap.(var)).(i);
        if i <> value then
          subtotal := logsumexp2 !subtotal imsg.(nmap.(var)).(i)
      end
    done;
    (* Recurse when var = value *)
    let left_accu = accu +. (imsg.(nmap.(var)).(value) -. !total) in
    if left_accu <> log_zero then begin
      let reset = set_condition conditions (true, var, value) in
      add_probs schema imsg omsg neighbors nmap left_accu conditions left;
      reset ()
    end;
    (* Recurse when var != value *)
    let right_accu = accu +. !subtotal -. !total in
      (* log(1.0 -. exp(imsg.(nmap.(var)).(value) -. !total)) in *)
    if right_accu <> log_zero then begin
      let reset = set_condition conditions (false, var, value) in
      add_probs schema imsg omsg neighbors nmap right_accu conditions right;
      reset ()
    end


(* Generate an array representing which values of each variable are legal *)
let condarray schema neighbors nmap f =
  let conditions = 
    Array.map (fun n -> Array.create schema.(n) true) neighbors in
  let apply_cond (sense,var,value) =
    for i = 0 to schema.(var) - 1 do
      if (i = value) <> sense then
        conditions.(nmap.(var)).(i) <- false
    done in
  List.iter apply_cond f.F.cond;
  conditions

(* Compute the probability of a set of conditions, represented
 * in an array form generated by the above function. *)
let condprob c msg =
  (* If all values are legal, return log 1 (which is zero) *)
  if Array.for_all identity c then 0.
  (* Otherwise, add up probs of all legal values *)
  else begin
    let total = ref log_zero in
    for i = 0 to Array.length c - 1 do
      if c.(i) then
        total := logsumexp2 !total msg.(i)
    done;
    !total
  end

(* Helper function for add_fsprobs, that works on a single feature. *)
let add_fprobs schema imsg omsg psum neighbors nmap f = 
  let wt = f.F.weight in
  let conditions = condarray schema neighbors nmap f in
  let p = Array.sumf (Array.map2 condprob conditions imsg) in

  (* For each involved variable... *) 
  for i = 0 to Array.length omsg - 1 do
    (* Divide prob by total marginal of satisfying states *)
    let total = ref log_zero in
    for value = 0 to Array.length imsg.(i) - 1 do
      if conditions.(i).(value) then 
        total := logsumexp2 !total imsg.(i).(value)
    done;
    let p = p -. !total in

    (* For each value that satisfies feature,
       add wt*prob to marginal, and prob to total. *)
    for value = 0 to Array.length imsg.(i) - 1 do
      if conditions.(i).(value) then begin
        omsg.(i).(value) <- 
          logsumexp2 omsg.(i).(value) (wt +. p +. imsg.(i).(value));
        psum.(i).(value) <- logsumexp2 psum.(i).(value) p;
      end
    done
  done

(* Update messages for a potential that represents a set of
 * mutually exclusive features. *)
let add_fsprobs schema imsg omsg neighbors nmap fl =
  let psum = Array.map Array.copy omsg in
  List.iter (add_fprobs schema imsg omsg psum neighbors nmap) fl;
  for i = 0 to Array.length omsg - 1 do
    for value = 0 to Array.length omsg.(i) - 1 do
      (* If the probabilities of all relevant rules added up to
         less than 1, then add 1 times the remaining probability,
         to fill in the missing cases.  This is equivalent to
         saying that the value of the potential is 1 when none of
         the features is satisfied. *)
      if psum.(i).(value) < 0. then 
        let not_p = log(1. -. exp psum.(i).(value)) in
        omsg.(i).(value) <- 
          logsumexp2 omsg.(i).(value) (not_p +. imsg.(i).(value))
    done
  done


(* Update the belief propagation messages in and out of a 
 * single potential 
 *)
let update_potential_messages schema imsg f =

  (* Create bidirectional mapping between variable and neighbor indices. *)
  let neighbors_l = F.vars f in
  let neighbors = Array.of_list neighbors_l in 
  let nmap = Array.create (Array.length schema) (-1) in
  Array.iteri (fun i n -> nmap.(n) <- i) neighbors;
  let omsg = Array.map (fun i -> Array.create schema.(i) log_zero) neighbors in

  begin match f with
  | F.Tree root ->
    (* TODO: create conditions array only once and share it, for efficiency? *)
    let conditions = Array.map (fun dim -> Array.create dim true) schema in
    add_probs schema imsg omsg neighbors nmap 0.0 conditions root;

  | F.Table (vars,ranges,table) ->
    let procstate varstate =
      (* Compute product of potential and all incoming messages *)
      let p = (F.log_value f varstate) +. 
        Array.sumf (Array.map2 (fun i dist -> dist.(varstate.(i))) neighbors imsg) in
      (* Add probability to appropriate elemnt of matching outputs *)
      let add_var_prob i dist =
        dist.(varstate.(i)) <- logsumexp2 dist.(varstate.(i)) p in
      Array.iter2 add_var_prob neighbors omsg in
    Mn.Varstate.iter_state schema (Array.to_list neighbors) procstate
  | F.FeatureSet fl -> 
    add_fsprobs schema imsg omsg neighbors nmap fl
  | F.Feature feat -> 
    add_fsprobs schema imsg omsg neighbors nmap [feat]
  | F.Const _ -> ()
  end;
  for i = 0 to Array.length omsg - 1 do
    (* Factor out the incoming message *)
    for j = 0 to Array.length omsg.(i) - 1 do 
      if imsg.(i).(j) <> log_zero then 
        omsg.(i).(j) <- omsg.(i).(j) -. imsg.(i).(j)
    done;
    normalize_inplace_log omsg.(i)
  done;
  omsg

let pprod a1 a2 = Array.map2 ( +. ) a1 a2

let dist_divide a1 a2 =
  let a = Array.map2 (fun x1 x2 -> if x2 <> log_zero then x1 -. x2 else log_zero) a1 a2 in
  normalize_inplace_log a;
  a

let kld d1 d2 =
  let d1 = Array.map exp d1 in
  let d2 = Array.map exp d2 in
  Array.sumf (Array.map2 (fun a b -> a *. log(a /. b)) d1 d2)

let check_convergence threshold old_marg new_marg =
  let klds = Array.map2 kld old_marg new_marg in
  let maxkl = Array.max klds in
  dlogf "max kld = %f\n" maxkl;
  if maxkl < threshold then true else false

exception Converged

let run_bp max_iter threshold mn evidence = 
  let schema = Mn.schema mn in
  let numfact = Array.length (Mn.factors mn) in
  let evidence = Data.check_evidence schema evidence in
  let mn = Mn.simplify mn evidence in
  let allneighbors = 
    Array.map (fun f -> Array.of_list (F.vars f)) (Mn.factors mn) in
  let uniform dim = Array.create dim (log (1. /. (float_of_int dim))) in

  (* Create potential messages *)
  let gen_potential_msgs f  =
    let neighbors = Array.of_list (Mn.Factor.vars f) in
    Array.map (fun i -> uniform schema.(i)) neighbors in
  let potentials = Mn.factors mn in
  let potential_msgs = Array.map gen_potential_msgs potentials in
  
  (* Update repeatedly until convergence *)
  let marginals = ref (Array.map uniform schema) in
  begin try 
    for i = 1 to max_iter do

      (* Update messages using potentials *)
      dlogf "ITERATION %d:\n" i;
      let new_potentials = Array.map2 
        (update_potential_messages schema) potential_msgs potentials in
      Array.iter2 (fun m1 m2 -> Array.blit m1 0 m2 0 (Array.length m1)) 
        new_potentials potential_msgs;

      (* Update marginals *)
      let new_marginals = Array.map uniform schema in
      for i = 0 to numfact - 1 do 
        for j = 0 to Array.length allneighbors.(i) - 1 do
          new_marginals.(allneighbors.(i).(j)) <- 
            pprod new_marginals.(allneighbors.(i).(j)) potential_msgs.(i).(j)
        done
      done;
      Array.iter normalize_inplace_log new_marginals;

      (* Check for convergence *)
      if check_convergence threshold !marginals new_marginals then 
        raise Converged
      else 
        marginals := new_marginals;

      (* Send messages back to the potentials if we haven't converged yet. *)
      for i = 0 to numfact - 1 do 
        for j = 0 to Array.length allneighbors.(i) - 1 do
          (* Incoming message = variable marginal / outgoing message *)
          potential_msgs.(i).(j) <- 
            dist_divide !marginals.(allneighbors.(i).(j)) potential_msgs.(i).(j)
        done
      done;
    done
  with Converged -> () end;
  !marginals
