open Printf
open Ext
open MfImpl 


let big_int = 1000000000 (* = 1 billion *)

(* Globals used for command line parameters *)
let qfile = ref ""
let evfile = ref ""
let modelfile = ref ""
let mfile = ref ""
let marg_outfile = ref ""
let sameev = ref false
let max_iter = ref 50
let threshold = ref 0.0001

(* Currently disabled. *)
let sameinit = ref false
let smooth_init = ref 0.0001 

let usage = "Usage: mf -m <model> -mo <output> -ev <evidence> [...]"
let args = Arg.align
 ([("-m", Arg.Set_string modelfile, " Model file (XMOD, BIF, or MN)") ;
   ("-mo", Arg.Set_string marg_outfile, " Output marginals") ;
   ("-ev", Arg.Set_string evfile, " Evidence file");
   ("-sameev", Arg.Set sameev, " Use the same evidence for all queries");
   ("-q",  Arg.Set_string qfile,  " Query file");
   ("-maxiter", Arg.Set_int max_iter, " Maximum number of iterations [50]") ;
   ("-thresh", Arg.Set_float threshold, " Convergence threshold [0.0001]")]
 (* Reenable sometime in the future: 
   ("-init", Arg.Set_string mfile, " Initial variable marginals");
   ("-sameinit", Arg.Set sameinit, " Use the same initialization for all queries");
   ("-smoothinit", Arg.Set_float smooth_init, " Prior counts of uniform distribution for smoothing the initial marginal") *) 
   @ common_arguments)


let read_marginals schema mstream =
  let marg = Data.input_marginals mstream in
  let norm_weight = !smooth_init in
  let smooth dim x = 
    (x +. (norm_weight/.(float_of_int dim))) /. (1. +. norm_weight) in
  Array.map (Array.map2 smooth schema) marg 

let condition_marginals ev marginals =
  if Array.length ev = 0 then marginals
  else begin
    let cond_one e m = 
      if e < 0 then m
      else begin
        let m' = Array.create (Array.length m) 0. in
        m'.(e) <- 1.;
        m'
      end in
    Array.map2 cond_one ev marginals
  end


let create_features mn ev =
  start_time "features";
  let mn = Mn.simplify mn ev in
  let fa = Array.of_list (Mn.to_features mn) in
  dlogf "Feature construction time: %f\n" (elapsed "features");
  fa

let do_approx () =
  let mn = match filetype !modelfile with
    | BNFile -> Bn.to_mn (Bn.load_auto !modelfile)
    | MNFile -> Mn.load (open_in !modelfile)
    | _ -> nlogf "ERROR: Unsupported file type.\n"; exit (-1) in

  let schema = Mn.schema mn in

  (* Process arguments, open streams *)
  let evstream = if !evfile <> "" then open_in !evfile else stdin in 
  let qstream = if !qfile <> "" then open_in !qfile else stdin in 
  let has_mfile = !mfile <> "" in
  let mstream = if has_mfile then (open_in !mfile) else stdin in
  if not has_mfile then
    sameinit := true;
  if !evfile = "" then sameev := true;
  if !sameev then sameinit := true;

  (* Read in shared initialization for all queries *)
  let sharedinit =
    if !sameinit then
      if has_mfile then read_marginals schema mstream
      else [||] (* Empty array signifies "use the uniform distribution" *)
    else [||] in

  (* Run inference for a single set of evidence, then evaluate on
   * all of the queries (which are presumed to be compatible with that
   * evidence). *)
  let sharedmarg = 
    if !sameev then begin
      start_time "mf";
      let ev = if !evfile = "" then [||] else Data.input_example evstream in
      let ev = Data.check_evidence schema ev in
      let fa = create_features mn ev in
      let sm = MfImpl.meanfield !max_iter !threshold schema fa sharedinit in
      vlogf "MF time: %fs\n" (elapsed "mf");
      (ev, sm)
    end else ([||], [||]) in
  let mout = if !marg_outfile <> "" then open_out !marg_outfile 
             else log_stream log_normal in

  (* Log likelihood and timing statistics *)
  let sl = stats_make () in
  let st = stats_make () in

  let first_marginal = ref true in
  begin
  try while true do
    start_time "mf";
    let init = 
      if !sameinit then sharedinit else (read_marginals schema mstream) in
    let (ev, px) = 
      if !sameev then sharedmarg
      else 
        let ev = Data.input_example evstream in
        let ev = Data.check_evidence schema ev in
        let fa = create_features mn ev in 
        (ev, MfImpl.meanfield !max_iter !threshold schema fa init) in
    if !qfile <> "" then begin
      let q = Data.input_example qstream in
      begin try 
      Data.check_point schema q
      with Data.Invalid_length -> 
             printf "Error: Query contains wrong number of values.\n";
             exit (-1)
         | Data.Invalid_value -> 
             printf "Error: Query contains an out-of-bounds value.\n";
             exit (-1)
      end;
      (* Compute conditional log probability of query configuration *)
      let llvar var vmarg =
        if q.(var) >= 0 && ev.(var) < 0 then vmarg.(q.(var)) else 0. in
      let l = Array.sumf (Array.mapi llvar px) in
      let dt = elapsed "mf" in
      if log_exists log_verbose then
        vlogf "%f %f\n" l dt
      else
        nlogf "%f\n" l;
      stats_add sl l;
      stats_add st dt
    end;
    (* Write out marginals *)
    if !marg_outfile <> "" || !qfile = "" then begin
      (* Separate two marginal sets with a blank line *)
      if !first_marginal then
        first_marginal := false
      else
        output_string mout "\n";
      (* Print out marginal after conditioning on evidence *)
      let marginals = Array.map (Array.map exp) px in
      Data.output_marginals mout (condition_marginals ev marginals)
    end;
    (* Stop immediately if we only have a single piece of 
     * evidence and no queries. *)
    if !sameev && !qfile = "" then raise Data.Eof
  done with Data.Eof -> ()
  end;
  (sl, st)


let main () = 
  start_time "total";
  Arg.parse args ignore usage;
  if !modelfile = "" then
    (Arg.usage args usage; exit (-1));

  common_log_init();

  let (sl, st) = do_approx () in

  (* Print stats *)
  if stats_n sl > 0 then begin
    if classify_float (stats_stderr st) = FP_nan then
      vlogf "time = %f +/- 0\n" (stats_mean st)
    else
      vlogf "time = %f +/- %f\n" (stats_mean st) (stats_stderr st);
    nlogf "avg = %f +/- %f\n" (stats_mean sl) (stats_stderr sl);
  end;
  vlogf "Total time: %fs\n" (elapsed "total")

let _ = main ()
