open Ext
open Printf

module F = Mn.Factor

(* Create hashes that map variable/value pairs to the features in which 
 * they appear. *)
let make_fhash pfa =
  let no_hash = Hashtbl.create 100 in
  let yes_hash = Hashtbl.create 100 in
  let add_to_hash (lp, f) (sense, var, value) = 
    Hashtbl.add (if sense then yes_hash else no_hash) (var, value) (lp, f) in
  let hashf (lp, f) = 
    List.iter (add_to_hash (lp, f)) f.F.cond in
  Array.iter hashf pfa;
  (yes_hash, no_hash)

(* Normalize an array to represent a probability distribution *)
let normalize probs =
  (* First, pre-normalize by the max value, to avoid overflow/underflow *)
  let maxval = Array.max probs in
  let logz = log (Array.sumf_map (fun x -> exp(x -. maxval)) probs) in
  (* Then normalize by the sum of exponentiated values *)
  Array.map (fun x -> x -. maxval -. logz) probs 


(* Mean squared difference of the exponentiated values from two arrays.
 * Used for comparing probability distributions that are specified as
 * arrays of log probabilities. *)
let rmse x y =
  let sqdiff la lb = 
    let diff = exp la -. exp lb in 
    diff *. diff in
  let diffa = Array.map2 sqdiff x y in
  sqrt((Array.sumf diffa)/.(float_of_int (Array.length diffa)))


let rec meanfield_rec max_iter threshold schema (yh, nh) pfa marginals i =
  (* Compute probabilities of v_i != value_j for each i, j *)
  let neg_dist a =
    let total = Array.sumf_map exp a in
    Array.map (fun p -> log (total -. exp p)) a in
  let neg_marginals = Array.map neg_dist marginals in

  (* Update the marginal distribution of a single variable *)
  let max_change = ref 0.0 in
  let mf_update_var v =
    let var_value_sum i =
      let y_bfl = Hashtbl.find_all yh (v,i) in
      let n_bfl = Hashtbl.find_all nh (v,i) in
      let wparam_y (lp, f) = 
        f.F.weight *. (exp (!lp -. marginals.(v).(i))) in
      let wparam_n (lp, f) = 
        f.F.weight *. (exp (!lp -. neg_marginals.(v).(i))) in
      List.sumf_map wparam_y y_bfl -. List.sumf_map wparam_n n_bfl in
    let vmarg = normalize (Array.init schema.(v) var_value_sum) in
    let neg_vmarg = neg_dist vmarg in
    (* Note the difference from the last distribution *)
    let delta = rmse marginals.(v) vmarg in
    if delta > !max_change then max_change := delta;
    (* Update probs... *)
    for i = 0 to Array.length vmarg - 1 do
      let y_bfl = Hashtbl.find_all yh (v,i) in
      let minus_plus (lp, f) = 
        lp := !lp -. marginals.(v).(i) +. vmarg.(i) in
      List.iter minus_plus y_bfl ;
      let n_bfl = Hashtbl.find_all nh (v,i) in
      let plus_minus (lp, f) = 
        lp := !lp -. neg_marginals.(v).(i) +. neg_vmarg.(i) in
      List.iter plus_minus n_bfl 
    done;
    (* DEBUG *)
    (* printf "New dist for variable %d:" v; print_farray vmarg; *)
    marginals.(v) <- vmarg ;
    neg_marginals.(v) <- neg_vmarg in
  (* Update all variables *)
  for v = 0 to Array.length marginals - 1 do
    mf_update_var v
  done ;

  (* Continue until convergence *)
  if !max_change > threshold && i < max_iter then
    meanfield_rec max_iter threshold schema (yh, nh) pfa marginals (i+1)


let meanfield max_iter threshold schema fa init_marginals =
  start_time "meanfield";
  (* Copy initial marginals if present; 
   * otherwise, create a uniform initial distribution *)
  let marginals = 
    if Array.length init_marginals > 0 then
      Array.map (Array.map log) init_marginals 
    else
      let uniform_dist dim = 
        Array.create dim (log (1.0 /. float_of_int dim)) in
      Array.map uniform_dist schema in
  (* Map each feature to a tuple.  The first item in the tuple, lp,
   * holds the log probability of the feature, according to the 
   * current marginal distribution. The second is the original
   * feature. *)
  let get_bprob f = 
    let cond_to_wt (sense, var, value) =
      if sense then marginals.(var).(value)
      else log(1. -. exp marginals.(var).(value)) in
    let logprob_f = List.sumf_map cond_to_wt f.F.cond in
    let lp = ref logprob_f in
    (lp, f) in
  let pfa = Array.map get_bprob fa in
  (* Create hashes for finding features quickly *)
  let (yh, nh) = make_fhash pfa in
  meanfield_rec max_iter threshold schema (yh, nh) pfa marginals 1;
  dlogf "Opt time: %f\n" (elapsed "meanfield"); 
  marginals
